#include <fstream>
#include <string>
#include <cstring>

#define STRICT
#include <windows.h>

#include "tas.hpp"
#include "game.hpp"
#include "movie.hpp"
#include "cmd.hpp"
#include "ui.hpp"
#include "util.hpp"
#include "types.hpp"

#include "resource.h"

using namespace std;

namespace{
    const ulong MY_PATH_MAX = 1030;
    const ulong FRAME_WAIT_TURBO = 5;
    const char STATE_PREFIX[] = "tasryoko-";

    void inputToggle(HWND hWnd, Movie::Input::Button btn)
    {
        if(TAS::getMovieMode() == TAS::MOVIE_PLAY) return;
        TAS::setInput(btn, TAS::getInput(btn)==Movie::Input::RELEASE ? Movie::Input::PRESS
                                                                     : Movie::Input::RELEASE);
    }
    void inputRP(HWND hWnd, Movie::Input::Button btn)
    {
        if(TAS::getMovieMode() == TAS::MOVIE_PLAY) return;
        TAS::setInput(btn, Movie::Input::RELEASEPRESS);
    }

    bool getOFN_moviePlay(HWND hWnd, string& path)
    {
        char buf[MY_PATH_MAX] = "";
        OPENFILENAME ofn;
        memset(&ofn, 0, sizeof(ofn));
        ofn.lStructSize = sizeof(ofn);
        ofn.hwndOwner = hWnd;
        ofn.lpstrFilter = "rfr files(*.rfr)\0*.rfr\0" "all files(*.*)\0*.*\0\0";
        ofn.lpstrFile = buf;
        ofn.nMaxFile = NELEMS(buf);
        ofn.lpstrInitialDir = Game::getMovieDir().c_str();
        ofn.lpstrTitle = "Play movie";
        ofn.Flags = OFN_FILEMUSTEXIST;

        if(!GetOpenFileName(&ofn)) return false;
        path = buf;

        return true;
    }

    bool getSFN_movieRec(HWND hWnd, string& path)
    {
        char buf[MY_PATH_MAX] = "";
        OPENFILENAME ofn;
        memset(&ofn, 0, sizeof(ofn));
        ofn.lStructSize = sizeof(ofn);
        ofn.hwndOwner = hWnd;
        ofn.lpstrFilter = "rfr files(*.rfr)\0*.rfr\0" "all files(*.*)\0*.*\0\0";
        ofn.lpstrFile = buf;
        ofn.nMaxFile = NELEMS(buf);
        ofn.lpstrInitialDir = Game::getMovieDir().c_str();
        ofn.lpstrTitle = "Record movie";
        ofn.lpstrDefExt = "rfr";
        ofn.Flags = OFN_OVERWRITEPROMPT;

        if(!GetSaveFileName(&ofn)) return false;

        path = buf;
        return true;
    }

    string statePath(int n)
    {
        return Game::getMovieDir() + "\\" + STATE_PREFIX + util::to_dec(n) + ".rfr";
    }

    void stateLoad(HWND hWnd, int n)
    {
        Movie state(0, 0);
        if(!Movie::load(state, statePath(n))) goto FAIL;

        if(!TAS::stateLoad(state)) goto FAIL;
        UI::setStatusLine("State " + util::to_dec(n) + " loaded");
        UI::invalidate();
        return;

FAIL:
        UI::setStatusLine("State " + util::to_dec(n) + " load error");
        UI::invalidate();
    }

    void stateSave(HWND hWnd, int n)
    {
        if(!TAS::stateSave(statePath(n))) goto FAIL;

        UI::setStatusLine("State " + util::to_dec(n) + " saved");
        UI::invalidate();
        return;

FAIL:
        UI::setStatusLine("State " + util::to_dec(n) + " save error");
        UI::invalidate();
    }

    string getDlgItemText(HWND hDlg, int id)
    {
        char buf[1024];
        uint len = GetDlgItemText(hDlg, id, buf, NELEMS(buf));
        if(!len) return "";
        return string(buf, len);
    }

    void setDlgItemText(HWND hDlg, int id, const string& s)
    {
        SetDlgItemText(hDlg, id, s.c_str());
    }

    bool pref_dlgHandler_WM_INITDIALOG(HWND hDlg, WPARAM wp, LPARAM lp)
    {
        setDlgItemText(hDlg, IDC_EDIT_ADDRPOSX, util::to_hex(TAS::getAddrPosX(), 8, true));
        setDlgItemText(hDlg, IDC_EDIT_ADDRRAND, util::to_hex(TAS::getAddrRand(), 8, true));
        setDlgItemText(hDlg, IDC_EDIT_ADDRSTAGE, util::to_hex(TAS::getAddrStage(), 8, true));
        return true;
    }

    void pref_searchAddrPosX(HWND hDlg)
    {
        ulong addr;
        if(TAS::searchAddrPosX(addr))
            setDlgItemText(hDlg, IDC_EDIT_ADDRPOSX, util::to_hex(addr, 8, true));
    }

    void pref_apply(HWND hDlg)
    {
        ulong addr;

        string s = getDlgItemText(hDlg, IDC_EDIT_ADDRPOSX);
        if(util::to_ulong(s, addr, 16)) TAS::setAddrPosX(addr);

        s = getDlgItemText(hDlg, IDC_EDIT_ADDRRAND);
        if(util::to_ulong(s, addr, 16)) TAS::setAddrRand(addr);

        s = getDlgItemText(hDlg, IDC_EDIT_ADDRSTAGE);
        if(util::to_ulong(s, addr, 16)) TAS::setAddrStage(addr);
    }

    bool pref_dlgHandler_WM_COMMAND(HWND hDlg, WPARAM wp, LPARAM lp)
    {
        WORD id_ctl = LOWORD(wp);
        switch(id_ctl){
        case IDC_BUTTON_SEARCH_ADDRPOSX:
            pref_searchAddrPosX(hDlg);
            return true;
        case IDOK:
            pref_apply(hDlg);
            EndDialog(hDlg, 0);
            return true;
        case IDCANCEL:
            EndDialog(hDlg, 0);
            return true;
        }
        return false;
    }

    INT_PTR CALLBACK pref_dlgProc(HWND hDlg, UINT msg, WPARAM wp, LPARAM lp)
    {
        switch(msg){
#define MY_HANDLE_MSG(m) case m: if(pref_dlgHandler_##m (hDlg, wp, lp)) return TRUE; break
            MY_HANDLE_MSG(WM_INITDIALOG);
            MY_HANDLE_MSG(WM_COMMAND);
#undef MY_HANDLE_MSG
        }
        return FALSE;
    }
}

namespace Cmd{
    void inputLeftToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::LEFT); }
    void inputLeftRP(HWND hWnd) { inputRP(hWnd, Movie::Input::LEFT); }
    void inputRightToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::RIGHT); }
    void inputRightRP(HWND hWnd) { inputRP(hWnd, Movie::Input::RIGHT); }
    void inputUpToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::UP); }
    void inputUpRP(HWND hWnd) { inputRP(hWnd, Movie::Input::UP); }
    void inputDownToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::DOWN); }
    void inputDownRP(HWND hWnd) { inputRP(hWnd, Movie::Input::DOWN); }
    void inputAttackToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::ATTACK); }
    void inputAttackRP(HWND hWnd) { inputRP(hWnd, Movie::Input::ATTACK); }
    void inputJumpToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::JUMP); }
    void inputJumpRP(HWND hWnd) { inputRP(hWnd, Movie::Input::JUMP); }
    void inputCardToggle(HWND hWnd) { inputToggle(hWnd, Movie::Input::CARD); }
    void inputCardRP(HWND hWnd) { inputRP(hWnd, Movie::Input::CARD); }

    void frameadv(HWND hWnd)
    {
        if(TAS::isPaused())
            TAS::frameadv();
        else
            TAS::setPause(true);
    }
    void pauseToggle(HWND hWnd)
    {
        TAS::setPause(!TAS::isPaused());
    }
    void turboEnable(HWND hWnd) { TAS::setFrameWait(FRAME_WAIT_TURBO); }
    void turboDisable(HWND hWnd) { TAS::setFrameWait(TAS::FRAME_WAIT_DEF); }
    void turboToggle(HWND hWnd)
    {
        if(TAS::getFrameWait() == TAS::FRAME_WAIT_DEF)
            turboEnable(hWnd);
        else
            turboDisable(hWnd);
    }

    void moviePlay(HWND hWnd)
    {
        if(TAS::getMovieMode() == TAS::MOVIE_REC) return;

        string path;
        if(!getOFN_moviePlay(hWnd, path)) return;
        Movie movie(0, 0);
        if(!Movie::load(movie, path)) return;

        if(movie.getStage() != TAS::getStage()){
            MessageBox(hWnd, "Stage mismatch", "tasryoko", MB_OK | MB_ICONERROR);
            return;
        }

        TAS::setMoviePath(path);
        TAS::setMovie(movie);
        TAS::setMovieMode(TAS::MOVIE_PLAY);

        TAS::setPause(true);
        TAS::retryStage();
        TAS::setRand(movie.getSeed());
    }

    void movieRec(HWND hWnd)
    {
        switch(TAS::getMovieMode()){
        case TAS::MOVIE_REC: return;
        case TAS::MOVIE_PLAY:
            {
                Movie movie = TAS::getMovie();
                movie.truncate(TAS::getFrameCount());
                TAS::setMovie(movie);
                TAS::setMovieMode(TAS::MOVIE_REC);
            }
            break;
        case TAS::MOVIE_NONE:
            string path;
            if(!getSFN_movieRec(hWnd, path)) return;

            TAS::setPause(true);
            TAS::retryStage();

            Movie movie(TAS::getStage(), TAS::getRand());
            TAS::setMoviePath(path);
            TAS::setMovie(movie);
            TAS::setMovieMode(TAS::MOVIE_REC);
            break;
        }
    }

    void movieStop(HWND hWnd)
    {
        switch(TAS::getMovieMode()){
        case TAS::MOVIE_NONE: return;
        case TAS::MOVIE_PLAY:
            TAS::setMovieMode(TAS::MOVIE_NONE);
            return;
        case TAS::MOVIE_REC:
            Movie movie = TAS::getMovie();
            movie.setState(TAS::genMovieState());
            if(!Movie::save(movie, TAS::getMoviePath())) goto FAIL;
            TAS::setMovieMode(TAS::MOVIE_NONE);
            break;
FAIL:
            MessageBox(hWnd, "Can't save movie", "tasryoko", MB_OK | MB_ICONERROR);
            break;
        }
    }

    void stateLoad0(HWND hWnd) { stateLoad(hWnd, 0); }
    void stateLoad1(HWND hWnd) { stateLoad(hWnd, 1); }
    void stateLoad2(HWND hWnd) { stateLoad(hWnd, 2); }
    void stateLoad3(HWND hWnd) { stateLoad(hWnd, 3); }
    void stateLoad4(HWND hWnd) { stateLoad(hWnd, 4); }
    void stateLoad5(HWND hWnd) { stateLoad(hWnd, 5); }
    void stateLoad6(HWND hWnd) { stateLoad(hWnd, 6); }
    void stateLoad7(HWND hWnd) { stateLoad(hWnd, 7); }
    void stateLoad8(HWND hWnd) { stateLoad(hWnd, 8); }
    void stateLoad9(HWND hWnd) { stateLoad(hWnd, 9); }

    void stateSave0(HWND hWnd) { stateSave(hWnd, 0); }
    void stateSave1(HWND hWnd) { stateSave(hWnd, 1); }
    void stateSave2(HWND hWnd) { stateSave(hWnd, 2); }
    void stateSave3(HWND hWnd) { stateSave(hWnd, 3); }
    void stateSave4(HWND hWnd) { stateSave(hWnd, 4); }
    void stateSave5(HWND hWnd) { stateSave(hWnd, 5); }
    void stateSave6(HWND hWnd) { stateSave(hWnd, 6); }
    void stateSave7(HWND hWnd) { stateSave(hWnd, 7); }
    void stateSave8(HWND hWnd) { stateSave(hWnd, 8); }
    void stateSave9(HWND hWnd) { stateSave(hWnd, 9); }

    void pref(HWND hWnd)
    {
        DialogBox(TAS::getHInstDLL(), "PREF", hWnd, pref_dlgProc);
    }

    void toggleHook(HWND hWnd)
    {
        if(Game::isHooked()){
            bool st = Game::unsetHook();
            UI::setStatusLine(string("unsetHook() ") + (st ? "OK" : "FAILED"));
            UI::invalidate();
        }
        else{
            bool st = Game::setHook();
            UI::setStatusLine(string("setHook() ") + (st ? "OK" : "FAILED"));
            UI::invalidate();
        }
    }
}
