#include <string>

#define STRICT
#include <windows.h>

#include "tas.hpp"
#include "game.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    HWND hWnd_ = 0;
    HHOOK hHookKbd_ = 0;
    HHOOK hHookCbt_ = 0;
    string baseDir_;
    string movieDir_;

    BOOL CALLBACK enumWin(HWND hWnd, LPARAM lp)
    {
        DWORD pid = static_cast<DWORD>(lp);
        DWORD pid_win;
        GetWindowThreadProcessId(hWnd, &pid_win);
        if(pid == pid_win){
            hWnd_ = hWnd;
            return FALSE;
        }
        return TRUE;
    }

    bool findWin()
    {
        DWORD pid = GetCurrentProcessId();
        EnumWindows(enumWin, static_cast<LPARAM>(pid));
        return hWnd_;
    }

    bool getDir()
    {
        const int MY_PATH_MAX = 1030;
        char buf[MY_PATH_MAX];
        DWORD len;
        if(!(len = GetCurrentDirectory(NELEMS(buf), buf))) return false;
        baseDir_.assign(buf, len);
        movieDir_ = baseDir_ + "\\replay";
        return true;
    }

    /**
     * 単に入力を遮断するだけ。キー入力はDLL側で生成したウィンドウで受ける。
     * PostThreadMessage() または PostMessage() でDLLスレッドへメッセージを
     * 横流しする手も考えられるが、これをやるとなぜかゲーム側への入力を遮断
     * できなくなる(WM_KEYDOWN, WM_KEYUP が素通りしてしまう)のでやめた。
     */
    LRESULT CALLBACK kbdProc(int code, WPARAM wp, LPARAM lp)
    {
        if(code < 0 || code == HC_NOREMOVE) return CallNextHookEx(hHookKbd_, code, wp, lp);
        return TRUE;
    }

    LRESULT CALLBACK cbtProc(int code, WPARAM wp, LPARAM lp)
    {
        if(code != HCBT_ACTIVATE) return CallNextHookEx(hHookCbt_, code, wp, lp);
        return TRUE;
    }
}

namespace Game{
    bool init()
    {
        if(!findWin()) return false;
        if(!getDir()) return false;
        return true;
    }

    bool fin()
    {
        if(!unsetHook()) return false;
        return true;
    }

    bool isHooked()
    {
        return hHookKbd_ && hHookCbt_;
    }

    bool setHook()
    {
        /**
         * ムービー再生中にゲームウィンドウをアクティブ化するとdesyncするよ
         * うなので、キー入力だけでなくアクティブ化も阻止する
         */
        DWORD win_tid = GetWindowThreadProcessId(hWnd_, 0);
        if(!(hHookKbd_ = SetWindowsHookEx(WH_KEYBOARD, kbdProc, 0, win_tid))) return false;
        if(!(hHookCbt_ = SetWindowsHookEx(WH_CBT, cbtProc, 0, win_tid))) return false;
        return true;
    }

    bool unsetHook()
    {
        if(!hHookKbd_) return true;
        if(!UnhookWindowsHookEx(hHookKbd_) || !UnhookWindowsHookEx(hHookCbt_)) return false;
        hHookKbd_ = 0;
        hHookCbt_ = 0;
        return true;
    }

    bool readMem(ulong addr, char* buf, ulong size)
    {
        HANDLE hProc = GetCurrentProcess();
        ulong n_read;
        if(!ReadProcessMemory(hProc, reinterpret_cast<LPCVOID>(addr), buf, size, &n_read)) return false;
        if(n_read != size) return false;
        return true;
    }

    bool writeMem(ulong addr, const char* buf, ulong size)
    {
        HANDLE hProc = GetCurrentProcess();
        ulong n_written;
        if(!WriteProcessMemory(hProc, reinterpret_cast<LPVOID>(addr), buf, size, &n_written)) return false;
        if(n_written != size) return false;
        return true;
    }

    void sendKeyPress(uint key)
    {
        SendMessage(hWnd_, WM_KEYDOWN, static_cast<WPARAM>(key), 0);
    }

    void sendKeyRelease(uint key)
    {
        SendMessage(hWnd_, WM_KEYUP, static_cast<WPARAM>(key), 0);
    }

    void sendMouseClickL(uint x, uint y)
    {
        LPARAM lp = x | (y<<16);
        // WM_MOUSEMOVE を送らないと、メニューの位置がずれることがある
        SendMessage(hWnd_, WM_MOUSEMOVE, 0, lp);
        SendMessage(hWnd_, WM_LBUTTONDOWN, 0, lp);
        SendMessage(hWnd_, WM_LBUTTONUP, 0, lp);
    }

    void sendMouseClickR(uint x, uint y)
    {
        LPARAM lp = x | (y<<16);
        SendMessage(hWnd_, WM_MOUSEMOVE, 0, lp);
        SendMessage(hWnd_, WM_RBUTTONDOWN, 0, lp);
        SendMessage(hWnd_, WM_RBUTTONUP, 0, lp);
    }

    const string& getBaseDir()
    {
        return baseDir_;
    }

    const string& getMovieDir()
    {
        return movieDir_;
    }
}
