#define STRICT
#include <windows.h>

#include "cmd.hpp"
#include "hotkey.hpp"
#include "ui.hpp"
#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    struct KeyMap{
        uint mod;
        uint key;
        void (*cmd)(HWND hWnd);
    };
    const KeyMap KEYMAP[] = {
        { 0, 'A', Cmd::inputLeftToggle },
        { MOD_SHIFT, 'A', Cmd::inputLeftRP },
        { 0, 'D', Cmd::inputRightToggle },
        { MOD_SHIFT, 'D', Cmd::inputRightRP },
        { 0, 'W', Cmd::inputUpToggle },
        { MOD_SHIFT, 'W', Cmd::inputUpRP },
        { 0, 'S', Cmd::inputDownToggle },
        { MOD_SHIFT, 'S', Cmd::inputDownRP },
        { 0, 'Z', Cmd::inputAttackToggle },
        { MOD_SHIFT, 'Z', Cmd::inputAttackRP },
        { 0, 'X', Cmd::inputJumpToggle },
        { MOD_SHIFT, 'X', Cmd::inputJumpRP },
        { 0, 'C', Cmd::inputCardToggle },
        { MOD_SHIFT, 'C', Cmd::inputCardRP },

        { 0, VK_OEM_5, Cmd::frameadv },
        { 0, VK_PAUSE, Cmd::pauseToggle },
        { MOD_SHIFT, VK_TAB, Cmd::turboToggle },

        { MOD_ALT, 'P', Cmd::moviePlay },
        { MOD_ALT, 'R', Cmd::movieRec },
        { MOD_ALT, 'Q', Cmd::movieStop },

        { 0, VK_F1, Cmd::stateLoad0 },
        { 0, VK_F2, Cmd::stateLoad1 },
        { 0, VK_F3, Cmd::stateLoad2 },
        { 0, VK_F4, Cmd::stateLoad3 },
        { 0, VK_F5, Cmd::stateLoad4 },
        { 0, VK_F6, Cmd::stateLoad5 },
        { 0, VK_F7, Cmd::stateLoad6 },
        { 0, VK_F8, Cmd::stateLoad7 },
        { 0, VK_F9, Cmd::stateLoad8 },
        { 0, VK_F10, Cmd::stateLoad9 },

        { MOD_SHIFT, VK_F1, Cmd::stateSave0 },
        { MOD_SHIFT, VK_F2, Cmd::stateSave1 },
        { MOD_SHIFT, VK_F3, Cmd::stateSave2 },
        { MOD_SHIFT, VK_F4, Cmd::stateSave3 },
        { MOD_SHIFT, VK_F5, Cmd::stateSave4 },
        { MOD_SHIFT, VK_F6, Cmd::stateSave5 },
        { MOD_SHIFT, VK_F7, Cmd::stateSave6 },
        { MOD_SHIFT, VK_F8, Cmd::stateSave7 },
        { MOD_SHIFT, VK_F9, Cmd::stateSave8 },
        { MOD_SHIFT, VK_F10, Cmd::stateSave9 },

        { MOD_CONTROL, 'P', Cmd::pref },

        { MOD_CONTROL, 'H', Cmd::toggleHook },
    };

    uint getModState()
    {
        uint mod = 0;
        if(GetAsyncKeyState(VK_SHIFT) < 0) mod |= MOD_SHIFT;
        if(GetAsyncKeyState(VK_CONTROL) < 0) mod |= MOD_CONTROL;
        if(GetAsyncKeyState(VK_MENU) < 0) mod |= MOD_ALT;
        return mod;
    }
}

namespace Hotkey{
    bool press(HWND hWnd, uint key)
    {
        uint mod = getModState();
        if(key == VK_TAB && !mod){
            Cmd::turboEnable(hWnd);
            return true;
        }
        else{
            for(uint i = 0; i < NELEMS(KEYMAP); ++i){
                const KeyMap& km = KEYMAP[i];
                if(key == km.key && mod == km.mod){
                    km.cmd(hWnd);
                    UI::update();
                    return true;
                }
            }
        }
        return false;
    }

    bool release(HWND hWnd, uint key)
    {
        uint mod = getModState();
        if(key == VK_TAB && !mod){
            Cmd::turboDisable(hWnd);
            return true;
        }
        return false;
    }
}
