#ifndef TASRYOKO_MOVIE_HPP
#define TASRYOKO_MOVIE_HPP

#include <string>
#include <vector>
#include <map>

#include "types.hpp"

class Movie{
public:
    class Input{
    public:
        enum Button{ LEFT=0, RIGHT, UP, DOWN, ATTACK, JUMP, CARD };
        enum Op{ RELEASE = 0x00, PRESS = 0x01, RELEASEPRESS = 0x03 };
        static const Button BTN_FIRST = LEFT;
        static const Button BTN_LAST = CARD;
        Input();
        explicit Input(ulong value);
        explicit Input(const char* packed);
        Op get(Button btn) const;
        void set(Button btn, Op op);
        ulong value() const;
        void pack(char* buf) const;
    private:
        ulong value_;
    };

    class State{
    public:
        State(ulong rand, double x, double y, double vx, double vy);
        bool operator==(const State& that) const;
        bool operator!=(const State& that) const;
        friend class Movie;
    private:
        ulong rand_;
        double x_, y_, vx_, vy_;
    };
    static bool load(Movie& movie, const std::string& path_rfr);
    static bool save(const Movie& movie, const std::string& path_rfr);

    Movie(ulong stage, ulong seed);
    Movie(const Movie& that);

    ulong getStage() const;
    ulong getSeed() const;
    ulong getFrameCount() const;

    Input get(ulong frame) const;
    void append(const Input& input);
    void truncate(ulong frame);

    int getExtraWait(ulong frame) const;

    State getState() const;
    void setState(const State& state);

private:
    ulong stage_;
    ulong seed_;
    std::vector<Input> inputs_;
    std::map<ulong, int> extraWaits_;
    State state_;
};

Movie::Input::Button& operator++(Movie::Input::Button& btn);

#endif /* TASRYOKO_MOVIE_HPP */
