#define STRICT
#include <windows.h>
#include <mmsystem.h>

#include "throttle.hpp"
#include "types.hpp"

using namespace std;

namespace{
    LONG frame_ = 0;
    volatile bool tgtflag_ = false;

    CRITICAL_SECTION* tgtc()
    {
        static bool ini = false;
        static CRITICAL_SECTION value;
        if(!ini){
            InitializeCriticalSection(&value);
            ini = true;
        }
        return &value;
    }

#if 0
    HANDLE tgtev()
    {
        static HANDLE value = 0;
        if(!value)
            value = CreateEvent(0, TRUE, FALSE, 0);
        return value;
    }
#endif

    ulong& uptimeBase()
    {
        static ulong value = 0;
        if(value == 0) value = timeGetTime();
        return value;
    }
}

namespace Throttle{
    ulong getUptime()
    {
        return uptimeBase() + FRAME_TIME*frame_;
    }

    void frameadv()
    {
        /**
         * InterlockedIncrement() はメモリバリアを生成するようなので、
         * frameadv() 呼び出し前後のコードについて実行順序の入れ替えが起こら
         * ないことを保証できる、と思う。これが必須かどうかはよくわからない
         * が、パフォーマンス的には問題なさそうなので一応やっておく。
         */
        InterlockedIncrement(&frame_);
    }

    void waitBusy(ulong ms)
    {
        ulong time_base = timeGetTime();
        while(timeGetTime() - time_base < ms) Sleep(0);
    }

    void waitSleep(ulong ms)
    {
        Sleep(ms);
    }

    void waitTGT(int n)
    {
        /**
         * イベント待ちよりもビジーウェイトの方が速い模様
         */
        for(int i = 0; i < n; ++i){
            EnterCriticalSection(tgtc());
            tgtflag_ = false;
#if 0
            ResetEvent(tgtev());
#endif
            LeaveCriticalSection(tgtc());
            while(!tgtflag_) waitSleep(0);
#if 0
            WaitForSingleObject(tgtev(), INFINITE);
#endif
        }
    }

    DWORD WINAPI my_timeGetTime()
    {
        EnterCriticalSection(tgtc());
        tgtflag_ = true;
#if 0
        SetEvent(tgtev());
#endif
        LeaveCriticalSection(tgtc());
        return Throttle::getUptime();
    }
}
