#include <iomanip>
#include <sstream>
#include <string>
#include <iterator>
#include <algorithm>
#include <functional>
#include <utility>

#include "util.hpp"
#include "types.hpp"

using namespace std;

namespace{
    struct ChCmpIAscii : public binary_function<char, char, int>{
        int operator()(char c1, char c2) const
        {
            char lc1 = util::tolower_ascii(c1);
            char lc2 = util::tolower_ascii(c2);
            if(lc1 < lc2) return -1;
            if(lc1 > lc2) return 1;
            return 0;
        }
    };
    int str_cmpi_ascii_impl(const string& s1, const string& s2)
    {
        typedef pair<string::const_iterator, string::const_iterator> PSCI;
        PSCI p = mismatch(s1.begin(), s1.end(), s2.begin(), not2(ChCmpIAscii()));
        if(p.first == s1.end()){
            if(p.second == s2.end()) return 0;
            else return -1;
        }
        return ChCmpIAscii()(*p.first, *p.second);
    }
}

namespace util{
    bool to_ulong(const string& s, ulong& n, int base)
    {
        return to_ulong(s.c_str(), n, base);
    }

    bool to_ulong(const char* s, ulong& n, int base)
    {
        char* end;
        n = strtoul(s, &end, base);
        if(n == 0 && *end != '\0') return false;
        return true;
    }

    string to_dec(ulong n)
    {
        ostringstream oss;
        oss << n;
        return oss.str();
    }

    string to_dec(double n, int prec)
    {
        ostringstream oss;
        oss << fixed << setprecision(prec) << n;
        return oss.str();
    }

    string to_hex(ulong n, int w, bool pad)
    {
        ostringstream oss;
        oss << hex << setw(w);
        if(pad) oss << setfill('0');
        oss << n;
        return oss.str();
    }

    char tolower_ascii(char c)
    {
        return 'A'<=c && c<='Z' ? c-'A'+'a' : c;
    }

    string str_lower_ascii(const string& s)
    {
        string lower;
        transform(s.begin(), s.end(), back_inserter(lower), tolower_ascii);
        return lower;
    }

    int str_cmpi_ascii(const string& s1, const string& s2)
    {
        if(s1.size() <= s2.size())
            return str_cmpi_ascii_impl(s1, s2);
        else
            return -str_cmpi_ascii_impl(s2, s1);
    }
}
