#ifndef TASRYOKO_UTIL_HPP
#define TASRYOKO_UTIL_HPP

#include <string>
#include <map>

#include "types.hpp"

#define NELEMS(array) (sizeof(array)/sizeof((array)[0]))

namespace util{
    /**
     * とてもいい加減な pack(), unpack()
     * どうせゲームと同じ環境で動かすのだから、エンディアンは考慮しない
     * 最適化レベル -O2 の場合は -fno-strict-aliasing が必要なので注意
     */
    template<typename T> void pack(char* buf, const T& value)
    {
        *(reinterpret_cast<T*>(buf)) = value;
    }
    template<typename T> T unpack(const char* buf)
    {
        return *(reinterpret_cast<const T*>(buf));
    }

    bool to_ulong(const std::string& s, ulong& n, int base = 10);
    bool to_ulong(const char* s, ulong& n, int base = 10);
    std::string to_dec(ulong n);
    std::string to_dec(double n, int prec);
    std::string to_hex(ulong n, int w, bool pad = false);

    char tolower_ascii(char c);
    std::string str_lower_ascii(const std::string& s);
    int str_cmpi_ascii(const std::string& s1, const std::string s2);

    template<typename K, typename V, typename Comp, typename Alloc>
    V map_get(const std::map<K,V,Comp,Alloc>& m, const K& k, const V& v)
    {
        typedef std::map<K,V,Comp,Alloc> Map;
        typename Map::const_iterator i = m.find(k);
        if(i != m.end())
            return i->second;
        else
            return v;
    }
}

#endif /* TASRYOKO_UTIL_HPP */
